# defines the df.cohorts DataFrame accessor and associated methods

from pandas_flavor import register_dataframe_accessor

from .cohort_metrics import CohortAnalysisException, CohortMetrics


@register_dataframe_accessor("cohorts")
class CohortAnalysisAccessor:

    """
    Register DataFrame.cohorts accessor

    Provides methods that faciliate the setting of particular
    DataFrames as the clickstream or cohorts attributes of a
    cohort_analysis.CohortMetrics instance

    """

    def __init__(self, df):
        self._df = df

    def __repr__(self):
        return "DataFrame.cohorts (accessor for cohort_analysis methods)"

    def __str__(self):
        return self.__repr__()

    def set_as_clickstream(
        self,
        metrics: CohortMetrics,
        timestamp_col: str = "timestamp",
        user_id_col: str = "user_id",
    ):

        """
        Register DataFrame.cohorts.set_as_clickstream

        Sets the clickstream attribute of a supplied CohortMetrics
        instance to be a copy of the DataFrame on which this
        method is called

        INPUTS:
            metrics - a cohort_analysis.metrics.CohortMetrics instance

        KEYWORDS:
            timestamp_col = "timestamp"
            user_id_col = "user_id"

            If the DataFrame has different names for these columns,
            that can be indicated using these keywords

        """

        clickstream = self._df.copy()

        initial_value = metrics.clickstream

        if timestamp_col != "timestamp":
            clickstream.loc[:, "timestamp"] = self._df.loc[:, timestamp_col]

        if user_id_col != "user_id":
            clickstream.loc[:, "user_id"] = self._df.loc[:, user_id_col]

        metrics.clickstream = clickstream

        try:
            metrics.verify_clickstream()
        except CohortAnalysisException as error:
            metrics.clickstream = initial_value  # reset
            raise error  # raise the exception

        metrics.clickstream = clickstream[["timestamp", "user_id"]]

    def set_as_cohorts(
        self,
        metrics: CohortMetrics,
        user_id_col: str = "user_id",
        reference_timestamp_col: str = "reference_timestamp",
    ):

        """
        Register DataFrame.cohorts.set_as_cohorts

        Sets the cohorts attribute of a supplied CohortMetrics
        instance to be a copy of the DataFrame on which this
        method is called

        INPUTS:
            metrics - a cohort_analysis.metrics.CohortMetrics instance

        KEYWORDS:
            user_id_col = "user_id"
            reference_timestamp_col = "reference_timestamp"

            If the DataFrame has different names for these columns,
            that can be indicated using these keywords

        """

        cohorts_copy = self._df.copy()

        initial_value = metrics.cohorts

        if user_id_col != "user_id":
            cohorts_copy.loc[:, "user_id"] = self._df.loc[:, user_id_col]
            cohorts_copy = cohorts_copy.drop(user_id_col, axis="columns")

        if reference_timestamp_col != "reference_timestamp":
            cohorts_copy.loc[:, "reference_timestamp"] = self._df.loc[
                :, reference_timestamp_col
            ]
            cohorts_copy = cohorts_copy.drop(reference_timestamp_col, axis="columns")

        metrics.cohorts = cohorts_copy

        try:
            metrics.verify_cohorts()
        except CohortAnalysisException as error:
            metrics.cohorts = initial_value  # reset
            raise error  # raise the exception



# cohort_analysis/decorators
# provides custom decorators to modify the behaviour of functions or methods

import functools

from .exceptions import CohortAnalysisException

from typing import Callable


def check_attributes_are_set(self):

    """
    check_attributes_are_set

    RAISES:
        CohortAnalysisException

        if either of the attributes .clickstream or .cohorts
        are not set on the input (and tells you which ones aren't set)

    """

    clickstream_missing = self.clickstream is None
    cohorts_missing = self.cohorts is None

    if clickstream_missing:
        if cohorts_missing:
            raise CohortAnalysisException(
                """
                The clickstream and cohorts attributes (DataFrames) need to be set
                on this instance of the CohortMetrics object.
                """
            )
        raise CohortAnalysisException(
            """
            The clickstream attribute (DataFrame) needs to be set on this instance
            of the CohortMetrics object
            """
        )

    if cohorts_missing:  # only reach this if .clickstream was set
        raise CohortAnalysisException(
            """
            The cohorts attribute (DataFrame) needs to be set on this instance
            of the CohortMetrics object
            """
        )


def check_attributes(method: Callable) -> Callable:

    """
    check_attributes

    Decorator. Returns a modified version of the input method that
    invokes check_attributes_are_set before invoking the input method

    """

    @functools.wraps(method)  # don't overwrite the method's docstring
    def decorated_method(self, *args, **kwargs):
        check_attributes_are_set(self)
        output = method(self, *args, **kwargs)
        return output

    return decorated_method  # return the souped-up method we just defined



# __init__.py
from .accessors import CohortAnalysisAccessor, CohortMetrics

# as a convenience, we provide cohort_analysis.metrics (an instance of
# cohort_analysis.metrics.CohortMetrics) on import; however the
# crucial .clickstream and .cohorts methods are initially not set.

metrics = CohortMetrics()

# We also provide new pandas.DataFrame methods via pandas_flavor.
# As a convenience, use
#   df1.cohorts.set_clickstream(cohort_analysis.metrics)
# and
#   df2.cohorts.set_cohorts(cohort_analysis.metrics)
# to set the attributes on cohort_analysis.metrics

__version__ = "0.0.1"
